/*
 * Copyright 2023, Exotic Silicon, all rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
 * conditions are met:
 *
 * 1. This software is licensed exclusively under this specific license text.  The license text may not be changed, and the software
 *    including modified versions may not be re-licensed under any other license text.
 * 2. Redistributions of source code must retain the above copyright notice, this list of conditions, and the following disclaimer.
 * 3. Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the following
 *    disclaimer in the documentation and/or other materials provided with the distribution.
 * 4. All advertising materials mentioning features or use of this software must display the following acknowledgement: This product
 *    includes software developed by Exotic Silicon.
 * 5. The name of Exotic Silicon must not be used to endorse or promote products derived from this software without specific prior
 *    written permission.
 * 6. Redistributions of modified versions of the source code must be clearly identified as having been modified from the original.
 * 7. Redistributions in binary form that have been created from modified versions of the source code must clearly state in the
 *    documentation and/or other materials provided with the distribution that the source code has been modified from the original.
 *
 * THIS SOFTWARE IS PROVIDED 'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL EXOTIC SILICON BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 * IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * Demonstration email filter implementing IP-based greylisting.
 *
 * This source code is supporting material for a programming tutorial.  For more information, please visit:
 *
 * https://research.exoticsilicon.com/articles/mail_filters
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <signal.h>
#include <sys/stat.h>
#include <stdio.h>
#include <fcntl.h>

#define DEBUG_OUT(x) { write (STDERR_FILENO, x, sizeof(x)-1); }

#define RAW_BUFFER_SIZE 65536
#define LINE_BUFFER_SIZE 4096

#define GREY_MIN 600
#define GREY_MAX 21600
#define WHITE_MAX 864000

#define MAX_SESSIONS 1024
#define MAX_TUPLES 8192
#define MAX_TUPLES_PER_IP 16

#define MAX_FIELDS 11

#define TUPLE_DATABASE "/home/greylist/filter_greydata"

struct st_pipeline {
	unsigned char * value[MAX_FIELDS];
	unsigned int len[MAX_FIELDS];
	unsigned int totlen[MAX_FIELDS];
	unsigned int lastfield;
	};

/*
 * The struct st_session IP address field can require a maximum of 47 bytes if a port is included:
 * 32 for IPv6 hex digits, 7 for colon separators, 2 for square brackets, 1 for another colon separator, and 5 for the port.
 *
 * [xxxx:xxxx:xxxx:xxxx:xxxx:xxxx:xxxx:xxxx]:12345
 */

struct st_session {
	unsigned char sid[16];
	unsigned char ip[48];
	unsigned char * helo;
	unsigned char * env_from;
	unsigned int ip_len;
	unsigned int helo_len;
	unsigned int env_from_len;
	};

struct st_tuple {
	unsigned char ip[47];
	unsigned char flag_ok;
	unsigned char * helo;
	unsigned char * env_from;
	unsigned char * env_to;
	unsigned int ip_len;
	unsigned int helo_len;
	unsigned int env_from_len;
	unsigned int env_to_len;
	time_t timestamp;
	};

volatile int flag_sigpipe;

/*
 * Parse the supplied line of input, and fill an array of pointers.
 * For each field: start of the field, length of the field, length to end of raw line, (including all subsequent fields).
 *
 * The last event specific or phase specific field is typically read as raw data, in which the pipe character is no longer considered a field delimiter.
 * For this purpose the length to end of raw line value is useful.
 */

int parse_pipeline(unsigned char * buffer_line_in, unsigned int buffer_line_in_len, struct st_pipeline * pipeline) {
unsigned int i;
unsigned int pipecount;
pipecount=0;

memset(pipeline, 0, sizeof(struct st_pipeline));

/*
 * We check for the first character being 0x00 even though this function doesn't otherwise parse null terminators.  Since the first field shouldn't
 * ever begin with a null, this would almost certainly indicate a programming error elsewhere sending a valid line length but an incorrect pointer.
 */
if (buffer_line_in_len==0 || *buffer_line_in==0) {
	return (1);
	}

pipeline->value[0]=buffer_line_in;
for (i=1; i<buffer_line_in_len && pipecount < MAX_FIELDS - 1; i++) {
	if (*(buffer_line_in+i)=='|') {
		pipeline->len[pipecount]=i-(pipeline->value[pipecount]-buffer_line_in);
		pipeline->totlen[pipecount]=buffer_line_in_len-(pipeline->value[pipecount]-buffer_line_in);
		pipecount++;
		pipeline->value[pipecount]=buffer_line_in+i+1;
		}
	}
pipeline->len[pipecount]=buffer_line_in_len-(pipeline->value[pipecount]-buffer_line_in);
pipeline->totlen[pipecount]=buffer_line_in_len-(pipeline->value[pipecount]-buffer_line_in);
pipeline->lastfield=pipecount;
return (0);
}

/*
 * Fill the supplied output buffer with one line from the input raw buffer.
 * Read data in to the raw buffer as required.
 * The trailing newline in the output buffer will be replaced with a NULL, making the returned output buffer NULL terminated.
 */
int line_in(unsigned char * line_buffer, unsigned int * len, unsigned char * raw_buffer, unsigned int * raw_buffer_readpos, unsigned int * raw_buffer_writepos)
{
int bytesin;
unsigned int outpos;

outpos=0;

while (1) {
	if (*raw_buffer_readpos==*raw_buffer_writepos) {
		*raw_buffer_readpos=0;
		bytesin=read(STDIN_FILENO, raw_buffer, RAW_BUFFER_SIZE);
		/*
		 * If we get end of file, or an error reading stdin, just exit.
		 */
		if (bytesin==0) {
			DEBUG_OUT ("Error - unexpected EOF on STDIN\n");
			return(1);
			}
		if (bytesin==-1 ) {
			DEBUG_OUT ("Error - got non-EOF error reading STDIN\n");
			return(1);
			}
		*raw_buffer_writepos = bytesin;
		}
	/*
	 * If we have already reached the end of the output buffer then exit, as this shouldn't happen during normal operation.
	 */
	if (outpos==LINE_BUFFER_SIZE) {
		DEBUG_OUT ("Error - buffer size exceeded\n");
		return(1);
		}
	/*
	 * Read a character from the global buffer, and put it in the output buffer.  If it's 10, null terminate and return, else loop.
	 */
	if ((*(line_buffer+outpos++) = *(raw_buffer+(*raw_buffer_readpos)++)) == 0x0a) {
		*(line_buffer+outpos-1)=0;
		*len=outpos-1;
		return (0);
		}
	}
}

int tuple_add_update(struct st_tuple * tuple_table, unsigned int * tuples, struct st_session session, unsigned char * env_to, unsigned int env_to_len, int fd_l)
{
/*
 * Add or update an entry in the tuple table.
 * Values for IP, helo, and envelope from come from the session table.
 * The recipient address is supplied from the current pipeline.
 * We don't care about the port part of the source IP, so we strip it by setting this function's local copy of ip_len back to the last colon.
 * This works for both IPv4 and IPv6 addresses.
 */
unsigned int i;
unsigned int per_ip_count;
time_t now;
now=time(NULL);
for (i=session.ip_len-1; i>0; i--) {
	if (session.ip[i]==':') {
		session.ip_len=i;
		break;
		}
	}
/*
 * See if the supplied IP address is already whitelisted regardless of the helo, sender and recipient that whitelisted it.
 * If it is, update the timestamp and return.  This avoids creating large numbers of tuples for the same IP address.
 */
per_ip_count=0;
for (i=0; i<*(tuples); i++) {
	if (tuple_table[i].ip_len == session.ip_len && memcmp(tuple_table[i].ip, session.ip, tuple_table[i].ip_len) == 0) {
		if (tuple_table[i].flag_ok == 1 && now-tuple_table[i].timestamp <= WHITE_MAX) {
			tuple_table[i].timestamp=now;
			dprintf (fd_l, "Updated timestamp on already whitelisted tuple %d\n", i);
			return (0);
			}
		if (tuple_table[i].flag_ok == 0 && now-tuple_table[i].timestamp < GREY_MAX) {
			per_ip_count++;
			}
		}
	}
if (per_ip_count > (MAX_TUPLES_PER_IP / 2)) {
	dprintf (fd_l, "IP already has %d greylisted tuples\n", per_ip_count);
	}
for (i=0; i<*(tuples); i++) {
	if (tuple_table[i].ip_len == session.ip_len && tuple_table[i].helo_len == session.helo_len && tuple_table[i].env_from_len == session.env_from_len) {	
		if (tuple_table[i].env_to_len == env_to_len) {
			if (memcmp(tuple_table[i].ip, session.ip, tuple_table[i].ip_len) == 0 &&
			    memcmp(tuple_table[i].helo, session.helo, tuple_table[i].helo_len) == 0 &&
			    memcmp(tuple_table[i].env_from, session.env_from, tuple_table[i].env_from_len) == 0 &&
			    memcmp(tuple_table[i].env_to, env_to, tuple_table[i].env_to_len) == 0 ) {
				/*
				 * Matched existing tuple.
				 * If already whitelisted, then update the timestamp.
				 * If still greylisted, don't update the timestamp, but whitelist if it's more than GREY_MIN and less then GREY_MAX behind.
				 */
				if (tuple_table[i].flag_ok == 1) {
					if (now-tuple_table[i].timestamp > WHITE_MAX) {
						dprintf (fd_l, "Whitelisted tuple %d has exceeded WHITE_MAX, returning to greylisted status "
						    "with new current timestamp %lld\n", i, now);
						tuple_table[i].flag_ok=0;
						tuple_table[i].timestamp=now;
						return (0);
						}
					/*
					 * The following case should already have been handled by the first iteration over the tuples.
					 * However it's been included here in case future changes inadvertently change the program logic and
					 * allow this path to be followed.
					 */
					tuple_table[i].timestamp=now;
					dprintf (fd_l, "Updated timestamp on already whitelisted tuple %d, (unexpected code path!)\n", i);
					return (0);
					}
				/*
				 * Currently greylisted.  Update to whitelisted if > GREY_MIN and < GREY_MAX.
				 * If and only if it's now whitelisted, then also update the timestamp.
				 */
				if (now-tuple_table[i].timestamp > GREY_MIN && now-tuple_table[i].timestamp < GREY_MAX) {
					tuple_table[i].flag_ok=1;
					tuple_table[i].timestamp=now;
					dprintf (fd_l, "Whitelisted previously greylisted tuple %d\n", i);
					return (0);
					}
				/*
				 * If greylisted and >= GREY_MAX then update the timestamp as we effectively 'start again'.
				 */
				if (now-tuple_table[i].timestamp >= GREY_MAX) {
					dprintf (fd_l, "Matching tuple %d had passed GREY_MAX.  ", i);
					if (per_ip_count >= MAX_TUPLES_PER_IP) {
						dprintf (fd_l, "IP already has the configured maximum of %d greylisted tuples, ignoring\n", MAX_TUPLES_PER_IP);
						return (0);
						}
					dprintf (fd_l, "Resetting timestamp to current time and keeping as greylisted\n");
					tuple_table[i].timestamp=now;
					return (0);
					}
				/*
				 * Greylisted and hasn't yet reached GREY_MIN.  Do nothing.
				 */
				dprintf (fd_l, "Tuple %d is greylisted for %llu more seconds\n", i, GREY_MIN-(now-tuple_table[i].timestamp));
				return (0);
				}
			}
		}
	}
/*
 * Finished searching existing tuples, and didn't find a match.
 * Check to see if we can add a new entry for this IP without exceeding the per IP address limit.
 */
if (per_ip_count >= MAX_TUPLES_PER_IP) {
	dprintf (fd_l, "IP has already reached the configured maximum of %d greylisted tuples, silently ignoring\n", MAX_TUPLES_PER_IP);
	return (0);
	}
/*
 * Add a new entry if there is space.
 */
if (*(tuples) < MAX_TUPLES) {
	/*
	 * If there is space in the table then we just append a new entry.
	 */
	tuple_table[*tuples].ip_len=session.ip_len;
	memcpy(tuple_table[*tuples].ip, session.ip, session.ip_len);
	tuple_table[*tuples].helo_len=session.helo_len;
	tuple_table[*tuples].helo=malloc_conceal(session.helo_len);
	/*
	 * If memory allocation fails, we don't explicitly free any previous allocations that we've just done because the calling function is
	 * expected to terminate the program anyway upon receiving the return value of two.
	 */
	if (tuple_table[*tuples].helo == NULL) {
		return (2);
		}
	memcpy(tuple_table[*tuples].helo, session.helo, session.helo_len);
	tuple_table[*tuples].env_from_len=session.env_from_len;
	tuple_table[*tuples].env_from=malloc_conceal(session.env_from_len);
	if (tuple_table[*tuples].env_from == NULL) {
		return (2);
		}
	memcpy(tuple_table[*tuples].env_from, session.env_from, session.env_from_len);
	tuple_table[*tuples].env_to_len=env_to_len;
	tuple_table[*tuples].env_to=malloc_conceal(env_to_len);
	if (tuple_table[*tuples].env_to == NULL) {
		return (2);
		}
	memcpy(tuple_table[*tuples].env_to, env_to, env_to_len);
	tuple_table[*tuples].timestamp=now;
	tuple_table[*tuples].flag_ok=0;
	dprintf (fd_l, "No matching tuple found in database, adding new tuple at slot %d\n", *tuples);
	(*tuples)++;
	return (0);
	}
/*
 * The tuple table is full, so we need to find an entry to overwrite.
 * We look for an entry which is either:
 * 1. Currently greylisted but has exceeded GRAY_MAX and would therefore be treated as a new entry anyway.
 * 2. Currently whitelisted but has exceeded WHITE_MAX and would also therefore be treated as a new entry.
 */
for (i=0; i < *(tuples); i++) {
	if ( (tuple_table[i].flag_ok == 0 && now-tuple_table[i].timestamp >= GREY_MAX) ||
	    (tuple_table[i].flag_ok == 1 && now-tuple_table[i].timestamp > WHITE_MAX) ) {
		dprintf (fd_l, "Reached MAX_TUPLES when adding new entry - overwriting existing %slisted tuple at slot %d\n",
		    (tuple_table[i].flag_ok == 0 ? "grey" : "white"), i);
		tuple_table[i].ip_len=session.ip_len;
		memcpy(tuple_table[i].ip, session.ip, session.ip_len);
		tuple_table[i].helo_len=session.helo_len;
		free (tuple_table[i].helo);
		tuple_table[i].helo=malloc_conceal(session.helo_len);
		if (tuple_table[i].helo == NULL) {
			return (2);
			}
		memcpy(tuple_table[i].helo, session.helo, session.helo_len);
		tuple_table[i].env_from_len=session.env_from_len;
		free (tuple_table[i].env_from);
		tuple_table[i].env_from=malloc_conceal(session.env_from_len);
		if (tuple_table[i].env_from == NULL) {
			return (2);
			}
		memcpy(tuple_table[i].env_from, session.env_from, session.env_from_len);
		tuple_table[i].env_to_len=env_to_len;
		free (tuple_table[i].env_to);
		tuple_table[i].env_to=malloc_conceal(env_to_len);
		if (tuple_table[i].env_to == NULL) {
			return (2);
			}
		memcpy(tuple_table[i].env_to, env_to, env_to_len);
		tuple_table[i].timestamp=now;
		tuple_table[i].flag_ok=0;
		return (0);
		}
	}
/*
 * If that didn't create any space then the value of MAX_TUPLES is set too low for the current workload, which is a configuration error.
 * In this case, the filter doesn't exit with an error but simply continues to run without adding an entry for the new tuple.
 */
dprintf (fd_l, "Error - MAX_TUPLES is too small\n");
return (1);
}

int session_get_add(struct st_session * session_table, unsigned int * sessions, unsigned char * sid, unsigned int * session_entry)
{
unsigned int n;
for (n=0; n<*(sessions); n++) {
	if (memcmp(sid, session_table[n].sid, 16)==0) {
		*session_entry=n;
		return (0);
		}
	}

/*
 * No existing entry matching the session ID and token was found, so we add a new one.
 */

if (*(sessions) == MAX_SESSIONS) {
	/*
	 * We should not reach MAX_SESSIONS in normal operation, if it is set higher than the maximum number of simultaneous sessions that smtpd
	 * itself would allow.  Since reaching MAX_SESSIONS here indicates either a code logic or configuration error, we return with the value
	 * two, which the calling function should interpret as an error condition and exit.
	 */
	return(2);
	}
memcpy(session_table[*sessions].sid, sid, 16);
/*
 * Set pointers to NULL so that we know whether to free them later or not.
 * Other values can be left wild.
 */
session_table[*sessions].helo=NULL;
session_table[*sessions].env_from=NULL;
*session_entry=*sessions;
(*sessions)++;

return (1);
}

int session_delete(struct st_session * session_table, unsigned int * sessions, unsigned int session_to_delete)
{
/*
 * Free pointers helo and env_from if they were previously malloc'ed.
 */
if (session_table[session_to_delete].helo != NULL) {
	free(session_table[session_to_delete].helo);
	}
if (session_table[session_to_delete].env_from != NULL) {
	free(session_table[session_to_delete].env_from);
	}
if (session_to_delete==((*sessions)-1)) {
	(*sessions)--;
	return (0);
	}
memmove(&session_table[session_to_delete], &session_table[session_to_delete+1], ((*sessions) - session_to_delete - 1) * sizeof(struct st_session));
(*sessions)--;
return (0);
}

void disp_st(struct st_session * table, int sessions, int fd_logfile)
{
int i;
dprintf (fd_logfile, "Listing %d sessions:\n", sessions);
for (i=0; i< sessions; i++) {
	dprintf (fd_logfile, "list %d %.16s\n",i, table[i].sid);
	}
return;
}

int display_tuple_table(struct st_tuple * tuple_table, unsigned int tuples, int fd_logfile)
{
unsigned int i;
unsigned int flag_expired;
unsigned char * ip;
unsigned char * helo;
unsigned char * env_from;
unsigned char * env_to;
time_t now;

ip=malloc(4096);
helo=malloc(4096);
env_from=malloc(4096);
env_to=malloc(4096);
now=time(NULL);

/*
 * The entries are not explicitly null-terminated, so we copy them to temporary buffers and terminate them before passing them to printf.
 */

for (i=0; i<tuples; i++) {
	memcpy (ip, tuple_table[i].ip, tuple_table[i].ip_len);
	*(ip+tuple_table[i].ip_len)=0;
	memcpy (helo, tuple_table[i].helo, tuple_table[i].helo_len);
	*(helo+tuple_table[i].helo_len)=0;
	memcpy (env_from, tuple_table[i].env_from, tuple_table[i].env_from_len);
	*(env_from+tuple_table[i].env_from_len)=0;
	memcpy (env_to, tuple_table[i].env_to, tuple_table[i].env_to_len);
	*(env_to+tuple_table[i].env_to_len)=0;
	if (tuple_table[i].flag_ok == 0) {
		flag_expired=(now-tuple_table[i].timestamp >= GREY_MAX ? 1 : 0);
		}
	if (tuple_table[i].flag_ok == 1) {
		flag_expired=(now-tuple_table[i].timestamp > WHITE_MAX ? 1 : 0);
		}
	dprintf (fd_logfile, "Tuple: %d, IP: %s, helo: %s, from: %s, to: %s, timestamp: %lld, flags: %s %sEXPIRED\n", i, ip, helo, env_from, env_to,
	    tuple_table[i].timestamp, (tuple_table[i].flag_ok == 1 ? "WHITE" : "GREY"), (flag_expired == 1 ? "" : "UN"));
	}
free(env_to);
free(env_from);
free(helo);
free(ip);
return (0);
}

int ip_check(struct st_tuple * tuple_table, unsigned int tuples, unsigned char * ip, unsigned int ip_len)
{
unsigned int i;
/*
 * Strip port from supplied IP.
 */
for (i=ip_len-1; i>0; i--) {
	if (ip[i]==':') {
		ip_len=i;
		break;
		}
	}
for (i=0; i < tuples; i++) {
	if (tuple_table[i].ip_len == ip_len && memcmp(tuple_table[i].ip, ip, ip_len)==0 && tuple_table[i].flag_ok==1) {
		tuple_table[i].timestamp=time(NULL);
		return (0);
		}
	}
return (1);
}

int write_tuples_to_disk(struct st_tuple * tuple_table, unsigned int tuples, int fd_database, int fd_logfile)
{
unsigned int i;

if (lseek (fd_database, 0, SEEK_SET) == -1) {
	dprintf (fd_logfile, "Seek failed on %s\n", TUPLE_DATABASE);
	return (1);
	}
if (ftruncate (fd_database, 0) == -1) {
	dprintf (fd_logfile, "Truncate failed on %s\n", TUPLE_DATABASE);
	return (1);
	}
write (fd_database, "*TUPLE_DATABASE*", 16);
dprintf (fd_logfile, "Writing %d tuples to %s\n", tuples, TUPLE_DATABASE);
for (i=0; i<tuples; i++) {
	write(fd_database, &(tuple_table[i].ip), 47);
	write(fd_database, &(tuple_table[i].flag_ok), 1);
	write(fd_database, &(tuple_table[i].ip_len), sizeof(unsigned int));
	write(fd_database, &(tuple_table[i].helo_len), sizeof(unsigned int));
	write(fd_database, &(tuple_table[i].env_from_len), sizeof(unsigned int));
	write(fd_database, &(tuple_table[i].env_to_len), sizeof(unsigned int));
	write(fd_database, &(tuple_table[i].timestamp), sizeof(time_t));
	write(fd_database, tuple_table[i].helo, tuple_table[i].helo_len);
	write(fd_database, tuple_table[i].env_from, tuple_table[i].env_from_len);
	write(fd_database, tuple_table[i].env_to, tuple_table[i].env_to_len);
	}
return (0);
}

int read_tuples_from_disk(struct st_tuple * tuple_table, unsigned int * tuples, int fd_database, int fd_logfile)
{
int totlen;
unsigned int flag_expired;
unsigned char * magic;
struct stat sb;
time_t now;

if (fd_database == -1) {
	dprintf (fd_logfile, "Open of %s failed\n", TUPLE_DATABASE);
	return (1);
	}
if (fstat(fd_database, &sb) == -1) {
	dprintf (fd_logfile, "Stat of %s failed\n", TUPLE_DATABASE);
	return (1);
	}
if (sb.st_size == 0) {
	dprintf (fd_logfile, "On-disk database %s is empty\n", TUPLE_DATABASE);
	return (0);
	}
if (sb.st_size < 16) {
	dprintf (fd_logfile, "Size of %s is less than 16 bytes\n", TUPLE_DATABASE);
	return (1);
	}
magic=malloc(16);
if (read (fd_database, magic, 16) != 16) {
	dprintf (fd_logfile, "Failed to read file magic\n");
	free(magic);
	return (1);
	}
if (memcmp("*TUPLE_DATABASE*", magic, 16) != 0) {
	dprintf (fd_logfile, "On-disk tuple database has bad file magic\n");
	free(magic);
	return (1);
	}
free(magic);
dprintf (fd_logfile, "On-disk tuple database has valid file magic\n");

totlen=sb.st_size-16;
now=time(NULL);
while (totlen != 0) {
	if (*tuples == MAX_TUPLES) {
		dprintf (fd_logfile, "On-disk tuple database exceeds current MAX_TUPLES of %d\n", MAX_TUPLES);
		return (0);
		}
	if (totlen < (47 + 1 + 4*sizeof(unsigned int) + sizeof(time_t))) {
		dprintf (fd_logfile, "Unexpected end of data within first block\n");
		return (0);
	}
	if (read(fd_database, &(tuple_table[*tuples].ip), 47) == -1) {
		goto error_handler;
		}
	if (read(fd_database, &(tuple_table[*tuples].flag_ok), 1) == -1) {
		goto error_handler;
		}
	if (read(fd_database, &(tuple_table[*tuples].ip_len), sizeof(unsigned int)) == -1) {
		goto error_handler;
		}
	if (read(fd_database, &(tuple_table[*tuples].helo_len), sizeof(unsigned int)) == -1) {
		goto error_handler;
		}
	if (read(fd_database, &(tuple_table[*tuples].env_from_len), sizeof(unsigned int)) == -1) {
		goto error_handler;
		}
	if (read(fd_database, &(tuple_table[*tuples].env_to_len), sizeof(unsigned int)) == -1) {
		goto error_handler;
		}
	if (read(fd_database, &(tuple_table[*tuples].timestamp), sizeof(time_t)) == -1) {
		goto error_handler;
		}
	if (tuple_table[*tuples].ip_len > 47 || tuple_table[*tuples].helo_len > 255 || tuple_table[*tuples].env_from_len > 2048 ||
	    tuple_table[*tuples].env_to_len > 2048) {
		dprintf (fd_logfile, "Invalid length argument in tuple database\n");
		return (0);
		}
	totlen -= (47 + 1 + 4*sizeof(unsigned int) + sizeof(time_t));
	if (totlen < tuple_table[*tuples].helo_len + tuple_table[*tuples].env_from_len + tuple_table[*tuples].env_to_len) {
		dprintf (fd_logfile, "Unexpected end of data within second block\n");
		return (0);
		}
	tuple_table[*tuples].helo=malloc(tuple_table[*tuples].helo_len);
	tuple_table[*tuples].env_from=malloc(tuple_table[*tuples].env_from_len);
	tuple_table[*tuples].env_to=malloc(tuple_table[*tuples].env_to_len);
	if (read(fd_database, tuple_table[*tuples].helo, tuple_table[*tuples].helo_len) == -1) {
		goto error_handler_with_free;
		}
	if (read(fd_database, tuple_table[*tuples].env_from, tuple_table[*tuples].env_from_len) == -1) {
		goto error_handler_with_free;
		}
	if (read(fd_database, tuple_table[*tuples].env_to, tuple_table[*tuples].env_to_len) == -1) {
		goto error_handler_with_free;
		}
	totlen -= (tuple_table[*tuples].helo_len + tuple_table[*tuples].env_from_len + tuple_table[*tuples].env_to_len);
	flag_expired=0;
	if (tuple_table[*tuples].flag_ok == 0 && (now-tuple_table[*tuples].timestamp >= GREY_MAX)) {
		flag_expired=1;
		}
	if (tuple_table[*tuples].flag_ok == 1 && (now-tuple_table[*tuples].timestamp > WHITE_MAX)) {
		flag_expired=1;
		}
	if (flag_expired==0) {
		(*tuples)++;
		}
	if (flag_expired==1) {
		free(tuple_table[*tuples].helo);
		free(tuple_table[*tuples].env_from);
		free(tuple_table[*tuples].env_to);
		}
	}
return (0);
error_handler_with_free:
free(tuple_table[*tuples].helo);
free(tuple_table[*tuples].env_from);
free(tuple_table[*tuples].env_to);
error_handler:
dprintf (fd_logfile, "Read call returned error\n");
return (0);
}

int ipv6_truncate(unsigned char * addr, unsigned int * len)
{
#define VALID_HEX(i) ((( i >= '0' && i <= '9') || (i >= 'a' && i <= 'f') || (i >= 'A' && i <= 'F')) ? 1 : 0)
#define VALUE_HEX(i) ((( i >= '0' && i <= '9') ? i - '0' : 0) + ((i >= 'a' && i <= 'f') ? i - 'a' + 10 : 0) + ((i >= 'A' && i <= 'F') ? i - 'A' + 10 : 0))
unsigned int i;
unsigned int colons;
unsigned int flag_double_colon;
unsigned int dcword;
unsigned int cword;
unsigned int words[8];
unsigned int hex_digits;
unsigned int specified;

colons=0;
flag_double_colon=0;
dcword=0;
cword=0;
hex_digits=0;
specified=0;

if (*len < 4 || *addr != '[') {
	return (1);
	}
/*
 * If the address starts with a colon it must be a double colon to be valid.
 */
if (*(addr+1)==':' && *(addr+2)!=':') {
	return (1);
	}
memset(words, 0, sizeof(words));
for (i=1; i<*len; i++) {
	if (cword==8) {
		return (1);
		}
	/*
	 * If it ends with a colon, it must have been a double colon to be valid.
	 */
	if (*(addr+i)==']') {
		if (hex_digits>0) {
			specified++;
			}
		if (*(addr+i-1)==':' && (i < 3 || *(addr+i-2)!=':')) {
			return (1);
			}
		break ;
		}
	if (!(VALID_HEX(*(addr+i))) && *(addr+i)!=':') {
		return (1);
		}
	if (VALID_HEX(*(addr+i))) {
		if (hex_digits==4) {
			return (1);
			}
		hex_digits++;
		words[cword] = words[cword] << 4;
		words[cword] |= VALUE_HEX(*(addr+i));
		}
	if (*(addr+i)==':') {
		if (*(addr+i-1)==':') {
			if (flag_double_colon==1) {
				return (1);
				}
			flag_double_colon=1;
			dcword=cword;
			cword--;
			}
		cword++;
		/*
		 * If we got at least one hex digit, then increase the count of actual specified blocks.
		 */
		if (hex_digits>0) {
			specified++;
			}
		hex_digits=0;
		}
	}
if (i==*len) {
	return (1);
	}
if (flag_double_colon==0 && cword != 7) {
	return (1);
	}
if (flag_double_colon==1 && specified==8) {
	return (1);
	}
#define SHIFT (7-cword)
#define TOSHIFT (cword-dcword+1)
if (flag_double_colon==1 && SHIFT != 0 && TOSHIFT != 0) {
	for (i=7; i >= dcword; i--) {
		words[i]=(i > 7 - TOSHIFT ? words[i-SHIFT] :0);
		}
	}
sprintf (addr, "%04x:%04x:%04x:%04x::0/64:", words[0], words[1], words[2], words[3]);
*len=26;
return (0);
}

void sigpipe()
{
flag_sigpipe=1;
return ;
}

int main()
{
struct st_session * session_table;
struct st_tuple * tuple_table;
unsigned char * raw_buffer;
unsigned char * line_buffer;
unsigned char * host_version;
unsigned int raw_buffer_readpos;
unsigned int raw_buffer_writepos;
unsigned int line_buffer_len;
unsigned int sessions;
unsigned int tuples;
unsigned int result;
unsigned int session_entry;
int fd_logfile;
int fd_database;
struct st_pipeline pipeline;

fd_logfile=open("/home/greylist/filter_greylog", O_WRONLY | O_CREAT | O_APPEND, 0600);
fd_database=open(TUPLE_DATABASE, O_RDWR | O_CREAT | O_EXLOCK, 0600);

pledge ("stdio", NULL);

raw_buffer_readpos=0;
raw_buffer_writepos=0;
raw_buffer=malloc_conceal(RAW_BUFFER_SIZE);
if (raw_buffer == NULL) {
	dprintf (fd_logfile, "Memory allocation error\n");
	return (1);
	}

line_buffer=malloc_conceal(LINE_BUFFER_SIZE);
if (line_buffer == NULL) {
	dprintf (fd_logfile, "Memory allocation error\n");
	return (1);
	}
line_buffer_len=0;

session_table=malloc_conceal(MAX_SESSIONS * sizeof(struct st_session));
if (session_table == NULL) {
	dprintf (fd_logfile, "Memory allocation error\n");
	return (1);
	}
sessions=0;

tuple_table=malloc_conceal(MAX_TUPLES * sizeof(struct st_tuple));
if (tuple_table == NULL) {
	dprintf (fd_logfile, "Memory allocation error\n");
	return (1);
	}
tuples=0;

signal(SIGPIPE, &sigpipe);

dprintf (fd_logfile, "Starting with MAX_SESSIONS = %u, (%lu Kb memory allocated), MAX_TUPLES = %u, (%lu Kb)\n", MAX_SESSIONS,
	    MAX_SESSIONS * sizeof(struct st_session) / 1024, MAX_TUPLES, MAX_TUPLES * sizeof(struct st_tuple) / 1024);

result=read_tuples_from_disk(tuple_table, &tuples, fd_database, fd_logfile);
if (result == 0) {
	dprintf (fd_logfile, "Loaded data for %d existing tuples from %s\n", tuples, TUPLE_DATABASE);
	display_tuple_table(tuple_table, tuples, fd_logfile);
	}
if (result == 1) {
	dprintf (fd_logfile, "Error whilst accessing %s, no tuples loaded\n", TUPLE_DATABASE);
	}

host_version=NULL;

while (line_buffer_len != 12 || memcmp("config|ready", line_buffer, 12) != 0) {
	if (line_in(line_buffer, &line_buffer_len, raw_buffer, &raw_buffer_readpos, &raw_buffer_writepos) != 0) {
		dprintf (fd_logfile, "I/O error on STDIN whilst waiting for config|ready\n");
		return (1);
		}
	if (line_buffer_len > 21 && memcmp("config|smtpd-version|", line_buffer, 21) == 0) {
		host_version=malloc(line_buffer_len-20);
		if (host_version == NULL) {
			dprintf (fd_logfile, "Memory allocation error\n");
			return (1);
			}
		memcpy(host_version, line_buffer+21, line_buffer_len-21);
		*(host_version+line_buffer_len-21)=0;
		}
	}

write (STDOUT_FILENO,"register|report|smtp-in|link-connect\n",37);
write (STDOUT_FILENO,"register|report|smtp-in|link-disconnect\n",40);
write (STDOUT_FILENO,"register|report|smtp-in|link-identify\n",38);
write (STDOUT_FILENO,"register|report|smtp-in|tx-mail\n",32);
write (STDOUT_FILENO,"register|report|smtp-in|tx-rcpt\n",32);
write (STDOUT_FILENO,"register|filter|smtp-in|data\n",29);
write (STDOUT_FILENO,"register|ready\n",15);

while (line_in(line_buffer, &line_buffer_len, raw_buffer, &raw_buffer_readpos, &raw_buffer_writepos)==0) {
	dprintf (fd_logfile, "Processing line: %s\n",line_buffer);
	parse_pipeline(line_buffer, line_buffer_len, &pipeline);
	if (pipeline.value[5]==NULL) {
		dprintf (fd_logfile, "Non-conformant data on STDIN\n");
		return (1);
		}
	result=session_get_add(session_table, &sessions, pipeline.value[5], &session_entry);
	if (result==2) {
		dprintf (fd_logfile, "Maximum number of concurrent sessions reached\n");
		return (1);
		}
	if (result==1) {
		if (pipeline.len[0] != 6 || pipeline.len[4] != 12) {
			dprintf (fd_logfile, "New session didn't start with link-connect report\n");
			return(1);
			}
		if (memcmp(pipeline.value[0], "report", 6) != 0 || memcmp(pipeline.value[4], "link-connect", 12) != 0) {
			dprintf (fd_logfile, "New session didn't start with link-connect report\n");
			return(1);
			}
		/*
		 * Store pipeline.value[8] as the session.ip.
		 * An IP address and port will always fit in 47 bytes.
		 * A local socket path might be longer than 47 bytes so we need to check for this.
		 * Since this filter is only intended to process remote IP-based connections, encountering a local socket path would likely indicate
		 * a configuration error, so we just exit.
		 */
		if (pipeline.len[8] > 47) {
			dprintf (fd_logfile, "Connection source is not a valid IPv4 or IPv6 address\n");
			return(1);
			}
		session_table[session_entry].ip_len=pipeline.len[8];
		memcpy (session_table[session_entry].ip, pipeline.value[8], pipeline.len[8]);
		ipv6_truncate(session_table[session_entry].ip, &session_table[session_entry].ip_len);
		}
	if (pipeline.len[0] == 6 && memcmp(pipeline.value[0], "report", 6)==0) {
		if (pipeline.len[4] == 13 && memcmp(pipeline.value[4], "link-identify", 13)==0) {
			/*
			 * Store the supplied HELO in the session helo pointer
			 */
			if (session_table[session_entry].helo != NULL) {
				dprintf (fd_logfile, "Duplicate helo for the same session\n");
				free (session_table[session_entry].helo);
				}
			session_table[session_entry].helo = malloc_conceal(pipeline.totlen[7]);
			if (session_table[session_entry].helo == NULL) {
				dprintf (fd_logfile, "Memory allocation error\n");
				return (1);
				}
			memcpy (session_table[session_entry].helo, pipeline.value[7], pipeline.totlen[7]);
			session_table[session_entry].helo_len = pipeline.totlen[7];
			}
		if (pipeline.len[4] == 7 && memcmp(pipeline.value[4], "tx-mail", 7)==0) {
			/*
			 * Store the supplied envelope FROM in the session env_from pointer
			 */
			if (session_table[session_entry].env_from != NULL) {
				dprintf (fd_logfile, "Duplicate envelope FROM for the same session\n");
				free (session_table[session_entry].env_from);
				}
			session_table[session_entry].env_from = malloc_conceal(pipeline.totlen[8]);
			if (session_table[session_entry].env_from == NULL) {
				dprintf (fd_logfile, "Memory allocation error\n");
				return (1);
				}
			memcpy (session_table[session_entry].env_from, pipeline.value[8], pipeline.totlen[8]);
			session_table[session_entry].env_from_len = pipeline.totlen[8];
			}
		if (pipeline.len[4] == 7 && memcmp(pipeline.value[4], "tx-rcpt", 7)==0) {
			/*
			 * Add or update an entry in the tuple table.
			 * Values for IP, helo, and envelope from come from the session table.
			 * The recipient address is supplied from the current pipeline.
			 */
			display_tuple_table(tuple_table, tuples, fd_logfile);
			result=tuple_add_update(tuple_table, &tuples, session_table[session_entry], pipeline.value[8], pipeline.totlen[8], fd_logfile);
			if (result==2) {
				dprintf (fd_logfile, "Memory allocation error in tuple_add_update\n");
				return (1);
				}
			}
		if (pipeline.len[4] == 15 && memcmp(pipeline.value[4], "link-disconnect", 15)==0) {
			/*
			 * Remove the current session from the sessions database.
			 */
			dprintf (fd_logfile, "Removing session from sessions table\n");
			session_delete(session_table, &sessions, session_entry);
			}
		}
	if (pipeline.len[0] == 6 && memcmp(pipeline.value[0], "filter", 6)==0) {
		if (pipeline.len[4] == 4 && memcmp(pipeline.value[4], "data", 4)==0) {
			result=ip_check(tuple_table, tuples, session_table[session_entry].ip, session_table[session_entry].ip_len);
			write (STDOUT_FILENO, "filter-result|", 14);
			write (STDOUT_FILENO, pipeline.value[5], 34);
			if (result==0) {
				write (STDOUT_FILENO, "proceed\n", 8);
				dprintf (fd_logfile, "Passing\n");
				}
			if (result==1) {
				write (STDOUT_FILENO, "reject|421 Service temporarily unavailable\n", 43);
				dprintf (fd_logfile, "Rejecting with 421\n");
				}
			}
		}
	}
if (flag_sigpipe==1) {
	dprintf (fd_logfile, "Got SIGPIPE\n");
	if (fd_database != -1) {
		result=write_tuples_to_disk(tuple_table, tuples, fd_database, fd_logfile);
		}
	}
close (fd_database);
close (fd_logfile);
return (1);
}
